package stacey;

import beast.core.Input;
import beast.core.Operator;
import beast.evolution.tree.Tree;
import beast.util.Randomizer;
import stacey.debugtune.Checks;
import stacey.util.*;

import java.util.ArrayList;
import java.util.List;

/**
 * Created by Graham Jones on 02/11/2014.
 */
public class FocusedNodeHeightScaler extends Operator {

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<Tree> smcTreeInput =
            new Input<Tree>("smcTree",
                    "The species tree or minimal clusters tree", Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<Tree>> geneTreesInput =
            new Input<List<Tree>>("geneTree",
                    "All gene trees",
                    new ArrayList<Tree>());

    private OperatorsSMCTree smcTree;
    private OperatorsGeneTree[] geneTrees;
    private GeneTreeScalingInfo [] geneTreeScalingInfos;
    private double [] sTreeLogHeights;
    private int [] sTreeDistances;
    private double [] sTreeWeights;

    private final boolean debugFlag = Boolean.valueOf(System.getProperty("stacey.debug"));
    private int numberofdebugchecks = 0;
    private final static int maxnumberofdebugchecks = 100000;



    @Override
    public double proposal() {

        if (smcTreeInput.get().getLeafNodeCount() < 4) {
            return Double.NEGATIVE_INFINITY;
        }

        update();
        if (debugFlag  &&  numberofdebugchecks < maxnumberofdebugchecks) {
            assert Checks.debugAllCompatible(smcTree, geneTrees);
        }
        double logHR = doFocusedScalerMove();
        reset();

        if (debugFlag  &&  numberofdebugchecks < maxnumberofdebugchecks) {
            update();

            assert Checks.debugAllCompatible(smcTree, geneTrees);
            numberofdebugchecks++;
            reset();
        }

        return logHR;
    }




    private void update() {
        smcTree.update();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].update();
        }
    }


    private void reset() {
        smcTree.reset();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].reset();
        }
    }


    @Override
    public void initAndValidate() throws Exception {
        //super.initAndValidate(); // ??
        /*if (smcTreeInput.get().getLeafNodeCount() < 4) {
            throw new Exception("FocusedNodeHeightScaler cannot be used if there are less than 4 minimal clusters.");
        }
        TODO this upsets Beauti */

        Bindings bindings = new Bindings(smcTreeInput.get().getTaxonset());
        smcTree = new OperatorsSMCTree(smcTreeInput.get(), bindings);
        // make the PIOGeneTree
        geneTrees = new OperatorsGeneTree[geneTreesInput.get().size()];
        for (int i = 0; i < geneTreesInput.get().size(); i++) {
            geneTrees[i] = new OperatorsGeneTree(geneTreesInput.get().get(i), bindings);
        }
        geneTreeScalingInfos = new GeneTreeScalingInfo[geneTrees.length];
        for (int j = 0; j < geneTrees.length; j++) {
            geneTreeScalingInfos[j] = new GeneTreeScalingInfo(geneTreesInput.get().get(j).getNodeCount());
        }
        sTreeLogHeights = new double[smcTreeInput.get().getNodeCount()];
        sTreeDistances = new int[smcTreeInput.get().getNodeCount()];
        sTreeWeights = new double[smcTreeInput.get().getNodeCount()];

    }



    private class GeneTreeScalingInfo {
        final double [] logHeights;
        final int [] distances;
        final double [] weights;


        // Constructor claims memory
        public GeneTreeScalingInfo(int nNodes) {
            logHeights = new double[nNodes];
            distances = new int[nNodes];
            weights = new double[nNodes];
        }

        public double[] getLogHeights() {
            return logHeights;
        }


        public int[] getDistances() {
            return distances;
        }

        public double[] getWeights() {
            return weights;
        }
    }



    double doFocusedScalerMove() {

        // TODO This is a fix for beast.core.State$Trie memory
        for (int j = 0; j < geneTrees.length; j++) {
            geneTreesInput.get().get(j).startEditing(null);
        }

        // Stage 1 - choose focal node and fill in info.
        // Fill in log(heights) and distances for smc tree nodes.
        // Choose focal node, get unions of itself and children.
        // Convert distances to weights for the smc tree.
        // Fill in log(heights) and distances for each gene tree.
        // Convert distances to weights for each gene tree's nodes
        FSMoveSMCNodeUnions fsNUnions = smcTree.setUpForFSMove(sTreeLogHeights, sTreeDistances);
        int rootDistance = sTreeDistances[sTreeDistances.length-1];
        for (int n = 0; n < sTreeDistances.length; n++) {
            sTreeWeights[n] = distanceToWeight(sTreeDistances[n], rootDistance);
        }
        for (int j = 0; j < geneTreeScalingInfos.length; j++) {
            GeneTreeScalingInfo jGTSI = geneTreeScalingInfos[j];
            geneTrees[j].fillInDistancesLogHeights(fsNUnions, jGTSI.getDistances(), jGTSI.getLogHeights());
            int [] jDistances = jGTSI.getDistances();
            double [] jWeights = jGTSI.getWeights();
            int jRootDistance = jDistances[jDistances.length-1];
            for (int n = 0; n < jDistances.length; n++) {
                jWeights[n] = distanceToWeight(jDistances[n], jRootDistance);
            }
        }

        // Stage 2 - find bounds.
        // Bounds from branches within smc tree.
        // Bounds from branches within gene trees and between smc tree and gene trees.
        double [] interval = new double[2];
        interval[0] = Double.NEGATIVE_INFINITY;
        interval[1] = Double.POSITIVE_INFINITY;
        smcTree.getFocusedScalerBounds(interval, sTreeLogHeights, sTreeWeights);
        FSMoveSMCNodeInfo[] fsNodeInfos = smcTree.getAllInternalFSMoveSNInfos(sTreeLogHeights, sTreeWeights);
        for (int j = 0; j < geneTreeScalingInfos.length; j++) {
            GeneTreeScalingInfo jGTSI = geneTreeScalingInfos[j];
            geneTrees[j].getFocusedScalerBounds(interval, fsNodeInfos, jGTSI.getLogHeights(), jGTSI.getWeights());
        }
        if (Double.isInfinite(interval[0])) {
            System.out.println("BUG 1 in FocusedNodeHeightScaler.doFocusedScalerMove()");
        }
        if (Double.isInfinite(interval[1])) {
            System.out.println("BUG 1 in FocusedNodeHeightScaler.doFocusedScalerMove()");
        }
        if (interval[0] >= 0.0) {
            System.out.println("BUG 1 in FocusedNodeHeightScaler.doFocusedScalerMove()");
        }
        if (interval[1] <= 0.0) {
            System.out.println("BUG 1 in FocusedNodeHeightScaler.doFocusedScalerMove()");
        }
        assert !Double.isInfinite(interval[0]);
        assert !Double.isInfinite(interval[1]);
        assert interval[0] < 0.0;
        assert interval[1] > 0.0;

        // Stage 3 - carry out the move
        double logHR = 0.0;
        double nearlyOne = 0.99999999;  // to avoid numerical errors which might make a branch slightly negative
        double logSF = Randomizer.uniform(nearlyOne*interval[0], nearlyOne*interval[1]);
        logHR += smcTree.fsMoveDoScale(logSF, sTreeWeights);
        for (int j = 0; j < geneTrees.length; j++) {
            GeneTreeScalingInfo jGTSI = geneTreeScalingInfos[j];
            logHR += geneTrees[j].fsMoveDoScale(logSF, jGTSI.getWeights());
        }
        return logHR;
    }






    private double distanceToWeight(int distance, int rootDistance) {
        double alpha = 8.0;
        if (distance >= rootDistance) {
            return 0.0;
        } else {
            double x = (double)(rootDistance - distance) / (double)rootDistance;
            return (Math.pow(alpha, x) - 1.0) / (alpha - 1.0);
        }
    }





}
