/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/

package stacey;

import beast.core.Description;
import beast.core.Input;
import beast.core.Operator;
import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import beast.util.Randomizer;
import stacey.debugtune.Checks;
import stacey.util.Bindings;
import stacey.util.NudgeMoveSMCNodeInfo;
import stacey.util.OperatorsGeneTree;
import stacey.util.OperatorsSMCTree;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;


/**
 * Created by Graham Jones on 19/08/2014.
 */

@Description("A move which changes a node height in species or minimal clusters tree, " +
             "and changes some node heights in the gene trees in a way which preserves all " +
             "topologies and usually maintains compatibility. <br/>")

// This is a multiple-trees operator. Is it a TreeOperator?
public class NodesNudge extends Operator {

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<Tree> smcTreeInput =
            new Input<Tree>("smcTree",
                    "The species tree or minimal clusters tree", Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<Tree>> geneTreesInput =
            new Input<List<Tree>>("geneTree",
                    "All gene trees",
                    new ArrayList<Tree>());

    public enum GtreeNodeCriterion {
        MIXED_WITH_PURE_CHILDREN,
        MIXED_WITH_A_PURE_CHILD,
        MIXED
    }
    private final static double [] gtreeNodeCriterionWeights = {1,1,1};





    private OperatorsSMCTree smcTree;
    private OperatorsGeneTree[] geneTrees;


    private int [] rejectCounts;
    private int [] acceptCounts;
    private double [] upTotals;
    private double [] downTotals;
    private int lastcriterion;
    private double lastheightchange;

    private final boolean debugFlag = Boolean.valueOf(System.getProperty("stacey.debug"));
    private int numberofdebugchecks = 0;
    private final static int maxnumberofdebugchecks = 100000;




    @Override
    public double proposal() {

        if (smcTreeInput.get().getLeafNodeCount() < 3) {
            return Double.NEGATIVE_INFINITY;
        }

        update();
        if (debugFlag  &&  numberofdebugchecks < maxnumberofdebugchecks) {
            assert Checks.debugAllCompatible(smcTree, geneTrees);
        }

        double logHR = doNodesNudgeMove();
        reset();

        if (debugFlag  &&  numberofdebugchecks < maxnumberofdebugchecks) {
            update();

            assert Checks.debugAllCompatible(smcTree, geneTrees);
            numberofdebugchecks++;
            reset();
        }
        return logHR;
    }


    private void update() {
        smcTree.update();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].update();
        }
    }


    private void reset() {
        smcTree.reset();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].reset();
        }
    }



    @Override
    public void initAndValidate() throws Exception {
        //super.initAndValidate(); // ??
        Bindings bindings = new Bindings(smcTreeInput.get().getTaxonset());
        smcTree = new OperatorsSMCTree(smcTreeInput.get(), bindings);
        // make the PIOGeneTree
        geneTrees = new OperatorsGeneTree[geneTreesInput.get().size()];
        for (int i = 0; i < geneTreesInput.get().size(); i++) {
            geneTrees[i] = new OperatorsGeneTree(geneTreesInput.get().get(i), bindings);
        }

        // debug and tuning stuff
        rejectCounts = new int[GtreeNodeCriterion.values().length];
        acceptCounts = new int[GtreeNodeCriterion.values().length];
        upTotals = new double[GtreeNodeCriterion.values().length];
        downTotals = new double[GtreeNodeCriterion.values().length];
    }


    @Override
    public void accept() {
        super.accept();
        acceptCounts[lastcriterion]++;
        if (lastheightchange > 0) {
            upTotals[lastcriterion] += lastheightchange;
        } else {
            downTotals[lastcriterion] += lastheightchange;
        }


    }


    @Override
    public void reject(final int reason) {
        super.reject(reason);
        rejectCounts[lastcriterion]++;

    }


    @Override
    public void storeToFile(final PrintWriter out) {
        super.storeToFile(out);
        if (debugFlag) {
            out.print("{id:\"" + getID() + "\"");
            for (int ch = 0; ch < GtreeNodeCriterion.values().length;ch++) {
                out.print(" " + GtreeNodeCriterion.values()[ch].toString() +
                                ", accept:" + acceptCounts[ch] +
                                ", reject:" + rejectCounts[ch] +
                                ", downTotal:" + downTotals[ch] +
                                ", upTotals:" + upTotals[ch]
                );
            }
            out.print("}");
        }
    }


    /*************************************************************************************************/



    private double doNodesNudgeMove() {

        // TODO This is a fix for beast.core.State$Trie memory
        for (int j = 0; j < geneTrees.length; j++) {
            geneTreesInput.get().get(j).startEditing(null);
        }
        // what type of move
        assert GtreeNodeCriterion.values().length == gtreeNodeCriterionWeights.length;
        int choice = Randomizer.randomChoicePDF(gtreeNodeCriterionWeights);
        lastcriterion = choice;
        GtreeNodeCriterion criterion = GtreeNodeCriterion.values()[choice];
        return doNormalNodesNudgeMove(criterion);
    }



    // For doNodesNudgeMove(). All Nodes Nudge Moves are currently normal
    private double doNormalNodesNudgeMove(GtreeNodeCriterion criterion) {
        double logHR = 0.0;
        double [] interval;
        ArrayList< ArrayList<Node>> geneNodesToNudge = new ArrayList< ArrayList<Node>>(geneTrees.length);

        // Choose the node in smcTree
        NudgeMoveSMCNodeInfo nani = smcTree.chooseNodeForNudge();

        // Collect all nodes in all gene tree that are to be nudged.
        for (int j = 0; j < geneTrees.length; j++) {
            ArrayList<Node> pgtnodes = geneTrees[j].nodesToNudge(nani, criterion);
            geneNodesToNudge.add(pgtnodes);
        }

        // get the constraints from gene trees
        interval = getMaxNodesNudgeIntervalForGeneTrees(geneNodesToNudge);

        // get the constraints from smcTree.
        interval[0] = Math.max(interval[0], nani.getLftHeight() - nani.getNodeHeight());
        interval[0] = Math.max(interval[0], nani.getRgtHeight() - nani.getNodeHeight());
        if (nani.getAncHeight() == Double.POSITIVE_INFINITY) {
            interval[1] = Math.min(interval[1], 0.1); // in case node is root and no other constraint
        } else {
            interval[1] = Math.min(interval[1], nani.getAncHeight() - nani.getNodeHeight());
        }
        assert interval[0] <= 0.0;
        assert interval[1] >= 0.0;

        // choose the amount to nudge by
        double dh = Randomizer.uniform(interval[0], interval[1]);
        logHR += smcTree.doNodeNudge(nani.getNodeNr(), dh);
        lastheightchange = dh;

        logHR += doNodesNudgeOnGeneTrees(dh, geneNodesToNudge);
        return logHR;
    }



    // For doNormalNodesNudgeMove()
    private double[] getMaxNodesNudgeIntervalForGeneTrees(ArrayList< ArrayList<Node>> toNudge) {
        double [] interval = new double[2];
        interval[0] = -Double.MAX_VALUE;
        interval[1] = Double.MAX_VALUE;
        for (int pgt = 0; pgt < geneTrees.length; pgt++) {
            double [] pgtint = getMaxNodesNudgeIntervalWithinOneGeneTree(toNudge.get(pgt));
            interval[0] = Math.max(interval[0], pgtint[0]);
            interval[1] = Math.min(interval[1], pgtint[1]);
        }
        return interval;
    }



    // For getMaxNodesNudgeIntervalForGeneTrees()
    private double[] getMaxNodesNudgeIntervalWithinOneGeneTree(ArrayList<Node> nodes) {
        ArrayList<ArrayList<Node>> conncpts = findConnectedComponents(nodes);
        double[] interval = new double[2];
        interval[0] = -Double.MAX_VALUE;
        interval[1] = Double.MAX_VALUE;
        for (ArrayList<Node> cpt : conncpts) {
            Node root = cpt.get(0); // findConnectedComponentFrom() puts root first
            Node anc = root.getParent();
            if (anc != null) {
                interval[1] = Math.min(interval[1], anc.getHeight() - root.getHeight());
            }
            for (Node node : cpt) {
                double nhgt = node.getHeight();
                Node ch0 = node.getChild(0);
                Node ch1 = node.getChild(1);
                if (!cpt.contains(ch0)) {
                    interval[0] = Math.max(interval[0], ch0.getHeight() - nhgt);
                }
                if (!cpt.contains(ch1)) {
                    interval[0] = Math.max(interval[0], ch1.getHeight() - nhgt);
                }
            }
        }
        return interval;
    }



    // For getMaxNodesNudgeIntervalWithinOneGeneTree()
    private ArrayList<ArrayList<Node>> findConnectedComponents(ArrayList<Node> nodes) {
        ArrayList<ArrayList<Node>> conncpts = new ArrayList<ArrayList<Node>>();
        HashMap<Node, Integer> node2Index = new HashMap<Node, Integer>(nodes.size());
        int[] idxs = new int[nodes.size()];
        for (int i = 0; i < nodes.size(); i++) {
            node2Index.put(nodes.get(i), i);
        }
        for (int i = 0; i < nodes.size(); i++) {
            idxs[i] = -1;
        }
        int idx = 0;
        for (Node x : nodes) {
            // if not dealt with x yet, trace back to root of its connected component
            if (idxs[node2Index.get(x)] < 0) {
                while (x.getParent() != null && node2Index.containsKey(x.getParent())) {
                    x = x.getParent();
                }
                // find component from root and add it
                ArrayList<Node> cpt = findConnectedComponentFrom(x, node2Index);
                conncpts.add(cpt);
                // mark nodes in component as done (I'm only using idxs as flags)
                for (Node c : cpt) {
                    if (node2Index.get(c) == null) {
                        System.out.println("BUG in findConnectedComponents()");
                    }
                    assert node2Index.get(c) != null;
                    idxs[node2Index.get(c)] = idx;
                }
                idx++;
            }
        }
        return conncpts;
    }



    // For findConnectedComponents()
    private ArrayList<Node> findConnectedComponentFrom(Node x, HashMap<Node, Integer> node2Index) {
        ArrayList<Node> cpt = new ArrayList<Node>();
        cpt.add(x); // Note root added first
        // recursion from x within the set of nodes
        if (x.getChildCount() > 0) {
            Node ch0 = x.getChild(0);
            Node ch1 = x.getChild(1);
            if (ch0 != null && node2Index.containsKey(ch0)) {
                ArrayList<Node> cpt0 = findConnectedComponentFrom(ch0, node2Index);
                cpt.addAll(cpt0);
            }
            if (ch1 != null && node2Index.containsKey(ch1)) {
                ArrayList<Node> cpt1 = findConnectedComponentFrom(ch1, node2Index);
                cpt.addAll(cpt1);
            }
        }
        return cpt;
    }




    // iterates over gene trees to change heights
    private double doNodesNudgeOnGeneTrees(double dh, ArrayList< ArrayList<Node>> toNudge) {
        double logHR = 0.0;
        for (int pgt = 0; pgt < geneTrees.length; pgt++) {
            logHR += nudgeNodesInListBy(toNudge.get(pgt), dh);
        }
        return logHR;
    }




    // For doNodesNudgeOnGeneTrees(). This one actually edits the BEAST tree
    private double nudgeNodesInListBy(ArrayList<Node> nodes, double dh) {
        double logHR = 0.0;
        for (Node node : nodes) {
            node.setHeight(node.getHeight() + dh);
        }
        return logHR;
    }
}




/************************** Some experiments. See OperatorsGeneTree too **************************/

/*    public enum GtreeNodeCriterion {
        MIXED_WITH_PURE_CHILDREN_AND_PARENT_IN_GROUP,
        MIXED_WITH_PURE_CHILDREN,
        MIXED_WITH_A_PURE_CHILD,
        MIXED,
        PURE_MIXED_PARENT,
        PURE_MIXED_PARENT_OR_MIXED_WITH_PURE_CHILDREN,
        PURE_MIXED_PARENT_OR_MIXED_WITH_A_PURE_CHILD,
        PURE_MIXED_PARENT_OR_MIXED,
    }
    private final static double [] gtreeNodeCriterionWeights = {0,1,1,1,0,1,1,1};*/